package jp.kirikiri.tvp2env;

import jp.kirikiri.tjs2.TJSException;
import jp.kirikiri.tvp2.msg.Message;
import jp.kirikiri.tvp2.visual.BaseTransHandler;
import jp.kirikiri.tvp2.visual.ScanLineProvider;
import jp.kirikiri.tvp2.visual.SimpleImageProvider;
import jp.kirikiri.tvp2.visual.SimpleOptionProvider;

public class UniversalTransHandlerProvider extends CrossFadeTransHandlerProvider {

	private static final String VAGUE = "vague";
	private static final String RULE = "rule";

	private String[] mStringValueRet;

	public UniversalTransHandlerProvider() {
		super();
		mStringValueRet = new String[1];
	}
	@Override
	public String getName() { return "universal"; }

	protected BaseTransHandler getTransitionObject( SimpleOptionProvider options,
			SimpleImageProvider imagepro, int layertype, int src1w, int src1h,
			int src2w, int src2h ) throws TJSException {

		// retrieve "time" option
		int er = options.getAsNumber( TIME, mLogValueRet );
		if( er < 0 ) Message.throwExceptionMessage( Message.SpecifyOption, TIME );
		long time = mLogValueRet[0];
		if( time < 2) time = 2; // too small time may cause problem

		// retrieve "vague" option
		er = options.getAsNumber( VAGUE, mLogValueRet );
		int vague = (int) mLogValueRet[0];
		if( er < 0 ) vague = 64;

		// retrieve "rule" option and load it as an image
		er = options.getAsString( RULE, mStringValueRet );
		if( er < 0 ) Message.throwExceptionMessage( Message.SpecifyOption, RULE );
		String rulename = mStringValueRet[0];
		mStringValueRet[0] = null;
		ScanLineProvider scpro = imagepro.loadImage( rulename, 8, 0x02ffffff, src1w, src1h );
		if( scpro == null ) Message.throwExceptionMessage( Message.CannotLoadRuleGraphic, rulename );

		return new UniversalTransHander( options, layertype, time, vague, scpro );
	}
}
